"""
This module implements the L-infinity data fitting approach using a 
semi-smooth Newton method and automatic parameter choice.
"""

__author__ = "Christian Clason <christian.clason@uni-graz.at>"
__date__ = "May 18, 2012"

import numpy as np
import scipy.linalg as la
import scipy.sparse as sp


def heat(n):
    """Test problem: inverse heat equation (adapted from ReguTools)"""
    h = 1./n
    t = np.linspace(h/2,1-h/2,n)
    k = h/(2*np.sqrt(np.pi))*t**(-1.5)*np.exp(-1/(4*t))
    r = np.zeros((n,1)); r[0] = k[0]
    A = la.toeplitz(k,r)
    x = (t<1./10)*(75*t**2) + \
        (t>=1./10)*(t<3./20)*(0.75+(20*t-2)*(3-20*t)) + \
        (t>=3./20)*(t<1./2)*(0.75*np.exp(-2*(20*t-3)))
    b = np.dot(A,x)
    return A,b,x,t


def solve_parameter_l2(K,yd,xe,alpha0):
    """Solve L2 fitting problem with best parameter"""
    KtK = np.dot(K.T,K)
    Kty = np.dot(K.T,yd)
    I = np.identity(K.shape[0])

    alrange = np.linspace(alpha0/100,alpha0,101)
    err_min = 1e99
    for altest in alrange:
        xtest = la.solve(KtK + altest*I, Kty)
        err   = la.norm(xtest-xe)
        if err < err_min:
            err_min = err
            alpha   = altest
            x       = xtest

    return x,alpha


def solve_linffitting(K,yd,alpha,verbose=True):
    """Solve L-infinity fitting problem (by semi-smooth Newton method)"""
    n = K.shape[0]  # problem size
    maxit = 20      # max iterations in semismooth Newton method
    I = np.identity(n)

    # initialize iterates
    Ap = Am = np.zeros(n)       # active sets for upper, lower bound
    Ap_old, Am_old = Ap, Am

    # continuation strategy
    for gamma in 10**np.arange(12):
        if verbose: print "Solving for gamma = %1.0e" % (gamma)

        # semismooth Newton iteration
        for it in xrange(maxit):
            As = np.diag((Ap+Am))
            # setup Newton step
            Lx = gamma*np.dot(K.T,(Ap+Am)*yd)
            Lc = gamma*np.dot((Am-Ap).T,yd)
            
            Lxx = alpha*I + gamma*np.dot(K.T,np.dot(As,K))
            Lxc = (gamma*np.dot(K.T,Am-Ap)).reshape((n,1))
            Lcc = (n + gamma*np.sum(Ap+Am)).reshape((1,1))

            H = np.bmat([[Lxx,Lxc],[Lxc.T,Lcc]])
            F = np.hstack([Lx,Lc])

            # solve for Newton step
            dx = la.solve(H,F)
            x,c = dx[:-1].squeeze(),dx[-1]
            
            # update active sets
            res = np.dot(K,x) - yd
            Ap = (res >  c).astype(float)
            Am = (res < -c).astype(float)

            # terminate Newton iteration if active sets no longer change
            change = (Ap-Ap_old)+(Am-Am_old)
            update = len(change[change.nonzero()])
            if verbose:
                print "Iteration %d: %d points changed in active set" \
                        % (it+1,update)
            if update == 0: break

            Ap_old,Am_old = Ap,Am

        # terminate continuation if Newton iteration converged in one step
        if it == 0: break

    return x,c


def solve_parameter_linf(K,yd,xe,sigma,alpha):
    """Compute regularization parameter (via balancing principle)"""
    n = K.shape[0]  # problem size
    tol = 1e-3      # terminate if relative change in parameter less
    maxit = 20      # max fixed point iterations

    print "Computing regularization parameter"
    for it in xrange(maxit):
        aold  = alpha
        x,c   = solve_linffitting(K,yd,alpha,verbose=False)
        alpha = sigma * c/(la.norm(x)**2/(2*n))
        err   = la.norm(x-xe)/np.sqrt(n)
        print "Iteration %d: alpha = %g, err = %g" %(it+1,alpha,err)
        if abs(aold-alpha)<tol*aold: break
    
    return x,c,alpha


if __name__ == '__main__':
    """Test script for L^\infty data fitting """
    import matplotlib.pyplot as pl
    import argparse
    
    parser = argparse.ArgumentParser(description='Test L-infinity fitting')
    parser.add_argument('-n', '--points', dest='n', type=int, 
                        required=True, help='problem size')
    parser.add_argument('-d', '--noiselevel', dest='d', type=float, 
                        required=True, help='noise level')
    parser.add_argument('-t', '--noisetype', dest='nstype', 
                        choices = ('uniform', 'quantized', 'checker'),
                        default='uniform', help='noise type (default: uniform)')
    args = parser.parse_args()

    # test problem: construct matrix, exact solution, data
    n = args.n              # problem size
    K, ye, xe, t = heat(n)

    # noisy data
    if args.nstype == 'uniform':
        nmax  = args.d*np.max(np.abs(ye))
        noise = nmax*2*(np.random.rand(n,)-0.5)
        yd    = ye + noise
    elif args.nstype == 'quantized':
        nmax  = args.d*(np.max(ye)-np.min(ye))
        yd    = np.round(ye/nmax)*nmax
    elif args.nstype == 'checker':
        noise = args.d*np.max(ye)*(-1)**np.arange(n)
        yd    = ye + noise

    # compare linf and l2 reconstructions
    xa,c,alpha = solve_parameter_linf(K,yd,xe,0.008,1e-1)
    x2,alpha2  = solve_parameter_l2(K,yd,xe,alpha)

    print "\nNoise level: delta_inf = %e (estimated: c = %e)\n" \
            % (la.norm(yd-ye,np.inf),c)
    print "Reconstruction error (Linf, alpha = %e): %0.3e"\
            % (alpha,la.norm(xa-xe)/np.sqrt(n))
    print "Reconstruction error (L2, alpha = %e):   %0.3e"\
            % (alpha2,la.norm(x2-xe)/np.sqrt(n))

    # plot results
    pl.plot(t,ye,t,yd)
    pl.legend(('exact','noisy'))

    pl.plot(t,xe,t,xa,t,x2, figure=pl.figure())
    pl.legend(('exact','L^infty','L^2'))
    pl.show()
