"""
This module implements the L-infinity data fitting approach using a 
semi-smooth Newton method and automatic parameter choice for the
solution of inverse source problems discretized by finite elements.
Besides NumPy and SciPy, it requires the 'dolfin' module from the
open source FEniCS Project (http://fenicsproject.org/)
"""

__author__ = "Christian Clason <christian.clason@uni-graz.at>"
__date__ = "May 18, 2012"

import numpy as np
import scipy.linalg as la
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve
from dolfin import *


def setup_problem(n):
    """Test problem: inverse source problem (FEM discretization)"""

    nu = 1          # diffusion coefficient
    b  = (-2,0)     # advection coefficient

    # setup grid, finite elements
    mesh = UnitSquare(n,n)
    xx = mesh.coordinates()
    V = FunctionSpace(mesh,'CG',1)
    u = TrialFunction(V)
    v = TestFunction(V)

    # advection-diffusion equation with homogeneous Dirichlet conditions
    a = Constant(nu)*inner(grad(u),grad(v))*dx + \
            inner(Constant(b),grad(u))*v*dx
    bc = DirichletBC(V, Constant(0.0), lambda x,on_boundary: on_boundary)
   
    xe = Function(V)    # true solution
    ye = Function(V)    # exact data

    xe.vector()[:] = ((np.abs(xx[:,0]-0.5)<1./6)*(np.abs(xx[:,1]-0.5)<1./6))
    solve(a==(xe*v*dx), ye, bc)

    # construct scipy matrix from bilinear form a, boundary condition bc
    parameters['linear_algebra_backend'] = 'uBLAS'
    def assemble_csr(a,bc):
        row,col,val = assemble(a,bcs=bc).data()
        return sp.csr_matrix((val,col,row))

    A  = assemble_csr(a,bc)             # stiffness matrix
    AT = assemble_csr(adjoint(a),bc)    # stiffness matrix (adjoint)
    M  = assemble_csr(u*v*dx,bc)        # mass matrix
   
    return A,AT,M,ye,xe


def solve_l2fitting(A,AT,M,yd,alpha):
    """Solve L2 fitting problem (via KKT system)"""
    H = sp.bmat([[M, AT],[A, -1/alpha*M]],format='csr')
    F = np.hstack([M*yd,0*yd])
    X = spsolve(H,F)
    y,p = np.split(X,2)
    x = p/alpha
    return x


def solve_parameter_l2(A,AT,M,yd,xe,alpha0):
    """Solve L2 fitting problem with best parameter"""
    alrange = np.linspace(alpha0/100,alpha0,101)
    err_min = 1e99
    xtest = Function(xe)
    for altest in alrange:
        xtest.vector()[:] = solve_l2fitting(A,AT,M,yd,altest)
        err = errornorm(xtest,xe)
        if err < err_min:
            err_min = err
            alpha   = altest
            x       = xtest.copy()
    return x,alpha


def solve_linffitting(A,AT,M,yd,alpha,verbose=True):
    """Solve L-infinity fitting problem (by semi-smooth Newton method)"""
    n = A.shape[0]  # problem size
    maxit = 20      # max iterations in semismooth Newton method

    # initialize iterates
    Ap = Am = np.zeros(n)       # active sets for upper, lower bound
    Ap_old, Am_old = Ap, Am

    # continuation strategy
    for gamma in 10**np.arange(12):
        if verbose: print "Solving for gamma = %1.0e" % gamma

        # semismooth Newton iteration
        for it in xrange(maxit):
            # setup Newton step
            Ly = gamma*(Ap+Am)*yd
            Lc = gamma*np.dot((Am-Ap).T,yd)
           
            Lyy = gamma*sp.spdiags(Ap+Am,0,n,n,format='csr')
            Lyc = gamma*(Am-Ap).reshape((n,1))
            Lcc = (1 + gamma*np.sum(Ap+Am))

            H = sp.bmat([[-M/alpha,A,None],[AT,Lyy,Lyc],[None,Lyc.T,Lcc]],
                         format='csr')
            F = np.hstack([0*Ly,Ly,Lc])
            
            # solve for Newton step
            dx  = spsolve(H,F)
            p,y = np.split(dx[:-1],2)
            c   = dx[-1]

            # update active sets
            Ap = (y-yd >  c).astype(float)
            Am = (y-yd < -c).astype(float)

            # terminate Newton iteration if active sets no longer change
            change = (Ap-Ap_old)+(Am-Am_old)
            update = len(change[change.nonzero()])
            if verbose:
                print "Iteration %d: %d points changed in active set" \
                        % (it+1,update)
            if update == 0: break

            Ap_old,Am_old = Ap,Am

        # terminate continuation if Newton iteration converged in one step
        if it == 0: break
    
    x = 1/alpha*p
    return x,c


def solve_parameter_linf(A,T,M,yd,xe,sigma,alpha):
    """Compute regularization parameter (via balancing principle)"""
    n = A.shape[0]  # problem size
    tol = 1e-3      # terminate if relative change in parameter less
    maxit = 20      # max fixed point iterations
    
    x = Function(xe)
    for it in xrange(maxit):
        aold  = alpha
        x.vector()[:],c  = solve_linffitting(A,AT,M,yd,alpha,verbose=False)
        alpha = sigma * c/(norm(x)**2/2)
        err   = errornorm(x,xe)
        print "Iteration %d: alpha = %g, err = %g" %(it+1,alpha,err)
        if abs(aold-alpha)<tol*aold: break

    return x,c,alpha


if __name__ == '__main__':
    """Test script for L^\infty data fitting"""
    import argparse

    parser = argparse.ArgumentParser(description='Test L-infinity fitting')
    parser.add_argument('-n', '--points', dest='n', type=int, 
                        required=True, help='problem size')
    parser.add_argument('-d', '--noiselevel', dest='d', type=float, 
                        required=True, help='noise level')
    parser.add_argument('-t', '--noisetype', dest='nstype', 
                        choices = ('uniform', 'quantized', 'checker'),
                        default='uniform', help='noise type (default: uniform)')
    args = parser.parse_args()

    # assemble matrices, construct exact solution, data
    A, AT, M, ye, xe = setup_problem(args.n)
    nel = A.shape[0]          # problem size

    # noisy data
    yd = Function(ye)
    yd_np = yd.vector().array()
    if args.nstype == 'uniform':
        nmax  = args.d*np.max(np.abs(yd_np))
        yd_np += nmax*2*(np.random.rand(nel,)-0.5) 
    elif args.nstype == 'quantized':
        nmax  = args.d*(np.max(yd_np)-np.min(yd_np))
        yd_np = np.round(yd_np/nmax)*nmax
    elif args.nstype == 'checker':
        noise = args.d*np.max(yd_np)*(-1)**np.arange(nel)
        yd_np += noise
    yd.vector()[:] = yd_np
    
    # compare linf and l2 reconstructions
    x,c,alpha = solve_parameter_linf(A,AT,M,yd_np,xe,0.001,1e-1)
    x2,alpha2 = solve_parameter_l2(A,AT,M,yd_np,xe,alpha)

    print "\nNoise level: delta_inf = %e (estimated: c = %e)\n" \
            % (norm(yd.vector()-ye.vector(),'linf'),c)
    print "Reconstruction error (Linf, alpha = %e): %0.3e"\
            % (alpha,errornorm(xe,x))
    print "Reconstruction error (L2, alpha = %e):   %0.3e"\
            % (alpha2,errornorm(xe,x2))

    # plot results
    plot(xe,title='true solution')
    plot(ye,title='exact data'); plot(yd,title='noisy data')
    plot(x,title='Linf reconstruction'); plot(x2,title='L2 reconstruction')
    interactive()
